/// <reference types="node" />

/**
 * Retrieve all dependencies for "filePath"
 *
 * @param { string } filePath
 * @param { 'browser' | 'node' } platform
 */
export function getDependencies(
  filePath: string,
  platform: 'browser' | 'node',
): Promise<Array<string>>;

/**
 * Factory for creating `Server` instances
 */
export function server(
  filePath?: string | Array<string>,
  options?: ServerOptions,
): Promise<Server>;

// src/_.d.ts
export interface Config {
  activePort: number;
  applicationLoaderURL: import('url').URL;
  brokenNamedExportsPackages: Record<string, Array<string>>;
  bundleDirMetaPath: string;
  bundleDirName: string;
  bundleDirPath: string;
  defaultPort: number;
  directories: Array<string>;
  dirPath: string;
  dvlpDirPath: string;
  electronEntryURL: import('url').URL;
  esbuildTargetByExtension: {
    [extension: string]: string;
  };
  extensionsByType: {
    [type: string]: Array<string>;
  };
  latency: number;
  maxAge: string;
  maxAgeLong: string;
  serverStartTimeout: number;
  testing: boolean;
  typesByExtension: {
    [extension: string]: ContentType;
  };
  version: string;
  versionDirPath: string;
}

export type ContentType = 'css' | 'html' | 'js';

export interface Entry {
  directories: Array<string>;
  isApp: boolean;
  isElectron: boolean;
  isSecure: boolean;
  isStatic: boolean;
  main: string | undefined;
}

type Http2ServerRequest = import('http2').Http2ServerRequest;
type Http2ServerResponse = import('http2').Http2ServerResponse;
type IncomingMessage = import('http').IncomingMessage;
type ServerResponse = import('http').ServerResponse;
type HttpServer = import('http').Server;
type HttpServerOptions = import('http').ServerOptions;
type Http2SecureServer = import('http2').Http2SecureServer;
type Http2SecureServerOptions = import('http2').SecureServerOptions;
type esbuild = {
  build(
    options: import('esbuild').BuildOptions & { write: false },
  ): Promise<import('esbuild').BuildResult>;
  transform(
    input: string,
    options?: import('esbuild').TransformOptions,
  ): Promise<import('esbuild').TransformResult>;
};

type Req = (IncomingMessage | Http2ServerRequest) & {
  filePath: string;
  type?: ContentType;
  url: string;
  params?: { [key: string]: string } | {};
};
type Res = (ServerResponse | Http2ServerResponse) & {
  bundled: boolean;
  encoding: string;
  metrics: Metrics;
  mocked: boolean;
  rerouted: boolean;
  transformed: boolean;
  unhandled: boolean;
  url: string;
  error?: Error;
};
type RequestHandler = (req: Req, res: Res) => void;
interface DestroyableHttpServer extends HttpServer {
  destroy(): void;
}

// src/application-host/_.d.ts
export interface ApplicationWorker {
  readonly activeProcess?: import('node:worker_threads').Worker;
  readonly origins: Set<string>;
  readonly isListening: boolean;
  /**
   * Add `filePaths` to file watcher
   */
  addWatchFiles(filePaths: string | Array<string>): void;
  /**
   * Send message to the application thread
   */
  sendMessage(message: string | object | number | boolean | bigint): void;
}

export interface ApplicationProcessWorkerData {
  hostOrigin: string;
  postMessage(msg: ApplicationWorkerMessage): void;
  main?: string;
  serializedMocks?: Array<SerializedMock>;
}

export type ApplicationHostMessage = { type: 'start'; main: string };
export type ApplicationLoaderMessage = {
  type: 'dependency';
  filePath: string;
};
export type ApplicationWorkerMessage =
  | { type: 'error'; error: string }
  | { type: 'listening'; origin: string }
  | { type: 'started' }
  | { type: 'watch'; filePath: string; mode: 'read' | 'write' };

export interface ApplicationWorkerPendingHandle {
  promise: Promise<{ body: string; href: string }>;
  resolve: (value: { body: string; href: string }) => void;
  reject: (value?: unknown) => void;
}

// src/electron-host/_.d.ts
export interface ElectronProcess {
  readonly activeThread?: import('node:child_process').ChildProcess;
  readonly origins: Set<string>;
  readonly isListening: boolean;
  /**
   * Add `filePaths` to file watcher
   */
  addWatchFiles(filePaths: string | Array<string>): void;
  /**
   * Send message to the electron process
   */
  sendMessage(message: string | object | number | boolean | bigint): void;
}

export interface ElectronProcessWorkerData {
  hostOrigin: string;
  main: string;
  postMessage(msg: ElectronProcessMessage): void;
  serializedMocks?: Array<SerializedMock>;
}

export type ElectronProcessMessage =
  | { type: 'started' }
  | {
      type: 'listening';
      origin: string;
    }
  | { type: 'watch'; filePath: string; mode: 'read' | 'write' };

// src/hooks/_.d.ts
export interface Hooks {
  /**
   * Bundle non-esm node_modules dependency requested by the browser.
   * This hook is run after file read.
   */
  onDependencyBundle?(
    id: string,
    filePath: string,
    fileContents: string,
    context: DependencyBundleHookContext,
  ): Promise<string | undefined> | string | undefined;
  /**
   * Transform file contents for file requested by the browser.
   * This hook is run after file read, and before any modifications by dvlp.
   */
  onTransform?(
    filePath: string,
    fileContents: string,
    context: TransformHookContext,
  ): Promise<string | undefined> | string | undefined;
  /**
   * Manually resolve import specifier.
   * This hook is run for each import statement.
   * If returns "false", import re-writing is skipped.
   * If returns "undefined", import specifier is re-written using default resolver.
   * If "context.isDynamic", also possible to return replacement for whole expression.
   */
  onResolveImport?(
    specifier: string,
    context: ResolveHookContext,
    defaultResolve: DefaultResolve,
  ): string | false | undefined;
  /**
   * Manually handle response for incoming server request.
   * If returns "true", further processing by dvlp will be aborted.
   */
  onRequest?(
    request: IncomingMessage | Http2ServerRequest,
    response: ServerResponse | Http2ServerResponse,
  ): Promise<boolean | undefined> | boolean | undefined;
  /**
   * Modify response body before sending to the browser.
   * This hook is run after all modifications by dvlp, and before sending to the browser.
   */
  onSend?(filePath: string, responseBody: string): string | undefined;
  /**
   * Manually resolve import specifiers for application server.
   * @see https://nodejs.org/api/esm.html#resolvespecifier-context-nextresolve
   */
  onServerResolve?(
    specifier: string,
    context: { conditions: Array<string>; parentURL?: string },
    nextResolve: NodeResolveLoaderHook,
  ): { format?: string; url: string };
  /**
   * Transform file contents for application server.
   * @see https://nodejs.org/api/esm.html#loadurl-context-nextload
   */
  onServerTransform?(
    filePath: string,
    context: { format?: string },
    nextLoad: NodeLoadLoaderHook,
  ): { format: string; source: string | SharedArrayBuffer | Uint8Array };
}

export interface DependencyBundleHookContext {
  esbuild: Pick<esbuild, 'build'>;
}

export interface TransformHookContext {
  client: {
    manufacturer?: string;
    name?: string;
    ua: string;
    version?: string;
  };
  esbuild: esbuild;
}

export interface ResolveHookContext {
  importer: string;
  isDynamic: boolean;
}

export type DefaultResolve = (specifier: string, importer: string) => string | undefined;

export type NodeResolveLoaderHook = (
  specifier: string,
  context: { conditions: Array<string>; parentURL?: string },
  nextResolve: NodeResolveLoaderHook,
) => { format?: string; url: string };

export type NodeLoadLoaderHook = (
  url: string,
  context: { format?: string },
  defaultLoad: NodeLoadLoaderHook,
) => { format: string; source: string | SharedArrayBuffer | Uint8Array };

// src/mock/_.d.ts
export interface Mocks {
  addResponse(
    req: string | MockRequest,
    res: MockResponse | MockResponseHandler,
    once?: boolean,
    onMock?: () => void,
  ): () => void;
  addPushEvents(
    stream: string | MockPushStream,
    events: MockPushEvent | Array<MockPushEvent>,
  ): () => void;
  load(filePaths: string | Array<string>): void;
  matchResponse(href: string, req?: Req, res?: Res): boolean | MockResponseData;
  matchPushEvent(
    stream: string | MockPushStream,
    name: string,
    push: (stream: string | PushStream, event: PushEvent) => void,
  ): boolean;
  hasMatch(
    reqOrMockData:
      | string
      | URL
      | { url: string }
      | MockResponseData
      | MockStreamData,
  ): boolean;
  remove(
    reqOrMockData:
      | string
      | URL
      | { url: string }
      | MockResponseData
      | MockStreamData,
  ): void;
  clear(): void;
  /** @deprecated */
  clean(): void;
}

export type MockResponseDataType = 'html' | 'file' | 'json';
export type MockStreamDataType = 'ws' | 'es';

export interface MockResponseData {
  url: URL;
  originRegex: RegExp;
  pathRegex: RegExp;
  paramsMatch: import('path-to-regexp').MatchFunction;
  searchParams: URLSearchParams;
  ignoreSearch: boolean;
  once: boolean;
  filePath: string;
  type: MockResponseDataType;
  response: MockResponse | MockResponseHandler;
  callback?: () => void;
}

export interface MockStreamEventData {
  name?: string;
  message: Buffer | string | Record<string, any>;
  options: MockPushEventOptions & {
    protocol?: string;
  };
}

export interface MockStreamData {
  url: URL;
  originRegex: RegExp;
  pathRegex: RegExp;
  paramsMatch: import('path-to-regexp').MatchFunction;
  searchParams: URLSearchParams;
  ignoreSearch: boolean;
  filePath: string;
  type: MockStreamDataType;
  protocol: string;
  events: { [name: string]: Array<MockStreamEventData> };
}

export interface MockRequest {
  url: string;
  filePath?: string;
  ignoreSearch?: boolean;
}

export type MockResponseHandler = (req: Req, res: Res) => void;

export interface MockResponse {
  body: string | Record<string, any>;
  hang?: boolean;
  headers?: Record<string, any>;
  error?: boolean;
  missing?: boolean;
  offline?: boolean;
  status?: number;
}

export interface MockResponseJSONSchema {
  request: MockRequest;
  response: MockResponse;
}

export interface MockPushEventJSONSchema {
  stream: MockPushStream;
  events: Array<MockPushEvent>;
}

export interface MockPushStream {
  url: string;
  type: string;
  filePath?: string;
  ignoreSearch?: boolean;
  protocol?: string;
}

export interface MockPushEventOptions {
  delay?: number;
  connect?: boolean;
  event?: string;
  id?: string;
  namespace?: string;
}

export interface MockPushEvent {
  name: string;
  message?: Buffer | string | Record<string, any>;
  sequence?: Array<MockPushEvent>;
  options?: MockPushEventOptions;
}

export interface SerializedMock {
  href: string;
  originRegex: string;
  pathRegex: string;
  search: string;
  ignoreSearch: boolean;
  events?: Array<string>;
}

export interface DeserializedMock {
  href: string;
  originRegex: RegExp;
  pathRegex: RegExp;
  search: URLSearchParams;
  ignoreSearch: boolean;
  events?: Array<string>;
}

// src/push-events/_.d.ts
export interface PushClient {
  on(event: string, callback: (event: { data: string }) => void): void;
  send(msg: Buffer | string, options?: PushEventOptions): void;
  removeAllListeners(): void;
  close(): void;
}

export interface PushStream {
  url: string;
  type: string;
}

export interface PushEvent {
  message: Buffer | string | Record<string, unknown>;
  options?: PushEventOptions;
}

export interface PushEventOptions {
  id?: string; // EventSource ID
  event?: string; // EventSource event OR Socket.IO event
  namespace?: string; // Socket.IO namespace
  protocol?: string; // Socket.IO protocol
}

// src/resolver/_.d.ts
export interface Package {
  browser?: Record<string, string>;
  env: 'browser' | 'node';
  exports?: string | Record<string, string | Record<string, string>>;
  exportsConditions: Array<string>;
  imports?: string | Record<string, string | Record<string, string>>;
  isProjectPackage: boolean;
  manifestPath: string;
  main?: string;
  name: string;
  path: string;
  paths: Array<string>;
  type: 'module' | 'commonjs' | undefined;
  version: string;
}

export type ResolveResult = { filePath: string; format: Package['type']; url?: string };

// src/server/_.d.ts
export interface ServerOptions {
  /**
   * The command-line arguments to pass to the application thread or electron process (default `[]`).
   */
  argv?: Array<string>;
  /**
   * The path or glob pattern containing ".crt" and ".key" files.
   * This enables secure https mode by proxying all requests through a secure server (default `''`).
   */
  certsPath?: string | Array<string>;
  /**
   * Additional directories to use for resolving file requests (default `[]`).
   */
  directories?: Array<string>;
  /**
   * Run file as electron.js entry file (default `false`).
   */
  electron?: boolean;
  /**
   * The path to a custom hooks registration file (default `''`).
   */
  hooksPath?: string;
  /**
   * The path(s) to load mock files from.
   */
  mockPath?: string | Array<string>;
  /**
   * Port to expose on `localhost`.
   * Will use `process.env.PORT` if not specified here (default `8080`).
   */
  port?: number;
  /**
   * Enable/disable browser reloading (default `true`).
   */
  reload?: boolean;
  /**
   * Disable/enable all logging (default `false`).
   */
  silent?: boolean;
  /**
   * Disable/enable verbose logging (default `false`).
   */
  verbose?: boolean;
}

export interface Server {
  /**
   * The entry config
   */
  readonly entry: Entry;
  /**
   * The listening state
   */
  readonly isListening: boolean;
  /**
   * The localhost origin
   */
  readonly origin: string;
  /**
   * The `Mocks` instance
   */
  readonly mocks: Mocks;
  /**
   * The localhost port number
   */
  readonly port: number;
  /**
   * The active application worker thread, if initialised
   */
  readonly applicationWorker?: ApplicationWorker;
  /**
   * The active electron process, if initialised
   */
  readonly electronProcess?: ElectronProcess;
  /**
   * Add `filePaths` to file watcher
   */
  addWatchFiles(filePaths: string | Array<string>): void;
  /**
   * Destroy server instance
   */
  destroy(): Promise<void>;
}

// src/test-server/_.d.ts
export interface TestServerOptions {
  /**
   * Enable/disable automatic dummy responses.
   * If unable to resolve a request to a local file or mock,
   * the server will respond with a dummy file of the appropriate type (default `true`).
   */
  autorespond?: boolean;
  /**
   * The amount of artificial latency to introduce (in `ms`) for responses (default `50`).
   */
  latency?: number;
  /**
   * The port to expose on `localhost`. Will use `process.env.PORT` if not specified here (default `8080`).
   */
  port?: number;
  /**
   * The subpath from `process.cwd()` to prepend to relative paths (default `''`).
   */
  webroot?: string;
}

export class TestServer {
  latency: number;
  port: number;
  mocks: Mocks;
  webroot: string;
  constructor(options?: TestServerOptions);
  /**
   * Load mock files at `filePath`
   */
  loadMockFiles(filePath: string | Array<string>): void;
  /**
   * Register mock `response` for `request`.
   * If `once`, mock will be unregistered after first use.
   * If `onMock`, callback when response is mocked
   */
  mockResponse(
    request: string | MockRequest,
    response: MockResponse | MockResponseHandler,
    once?: boolean,
    onMockCallback?: () => void,
  ): void;
  /**
   * Register mock push `events` for `stream`
   */
  mockPushEvents(
    stream: string | MockPushStream,
    events: MockPushEvent | Array<MockPushEvent>,
    onSendCallback?: (data: any) => void,
  ): void;
  /**
   * Push data to WebSocket/EventSource clients
   * A string passed as `event` will be handled as a named mock push event
   */
  pushEvent(stream: string | PushStream, event?: string | PushEvent): void;
  /**
   * Clear all mock data
   */
  clearMockFiles(): void;
  /**
   * Prevent process from exiting while this server is active
   */
  ref(): void;
  /**
   * Allow process to exit if this is the only active
   */
  unref(): void;
  /**
   * Destroy server instance
   */
  destroy(): Promise<void>;
}

// src/utils/_.d.ts
export class Metrics {
  events: Map<string, [number, number]>;
  constructor(res: Res);
  recordEvent(name: string): void;
  getEvent(name: string, formatted?: boolean): string | number;
}
export namespace Metrics {
  export enum EVENT_NAMES {
    bundle = 'bundle file',
    csp = 'inject CSP header',
    imports = 'rewrite imports',
    mock = 'mock response',
    response = 'response',
    scripts = 'inject HTML scripts',
    transpile = 'transpile file',
  }
}

export interface PatchResponseOptions {
  directories?: Array<string>;
  footerScript?: {
    string: string;
    url?: string;
  };
  headerScript?: {
    string: string;
    url?: string;
  };
  send?(filePath: string, responseBody: string): string | undefined;
  resolveImport?(
    specifier: string,
    context: ResolveHookContext,
    defaultResolve: DefaultResolve,
  ): string | false | undefined;
}

export interface FindOptions {
  directories?: Array<string>;
  type?: ContentType;
}

export interface Platform {
  manufacturer?: string;
  name?: string;
  os?: {
    architecture?: number;
    family?: string;
    version?: string;
  };
  ua: string;
  version?: string;
}

export type InterceptClientRequestCallback = (url: URL) => boolean;
export type InterceptFileAccessCallback = (
  filePath: string,
  mode: 'read' | 'write',
) => void;
export type InterceptCreateServerCallback = (origin: string) => void;

export interface Watcher {
  has(filePath: string): boolean;
  add(filePath: string | Array<string> | Set<string>): void;
  remove(filePath: string, permanent?: boolean): void;
  close(): void;
}

export interface RequestContext {
  assert: ImportAssertionType;
  dynamic: boolean;
  filePath?: string;
  href: string;
  imported: boolean;
  type?: ContentType;
}

export type ImportAssertionType = 'css' | 'json' | undefined;

